/* FreeTDS - Library of routines accessing Sybase and Microsoft databases
 * Copyright (C) 1998-1999  Brian Bruns
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <limits.h>
#include <assert.h>
#include <ctype.h>
#ifdef __DGUX__
#include <paths.h>
#endif
#ifdef __FreeBSD__
#include <sys/time.h>
#endif
#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#define PATH_MAX 255
#endif
#ifndef WIN32
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif
#include "tds.h"
#include "tdsutil.h"
#ifdef DMALLOC
#include <dmalloc.h>
#endif

static char  software_version[]   = "$Id: config.c,v 1.5 2003/04/10 05:42:34 frank Exp $";
static void *no_unused_var_warn[] = {software_version,
                                     no_unused_var_warn};


static void tds_config_login(TDSCONFIGINFO *config, TDSLOGIN *login);
static int tds_read_conf_file(char *server, TDSCONFIGINFO *config);
static void tds_read_interfaces(char *server, TDSCONFIGINFO *config);
static void tds_config_verstr(char *tdsver, TDSCONFIGINFO *config);
static int tds_config_boolean(char *value);
static void lookup_host(const char *servername, const char *portname, char *ip, char *port);
static int parse_server_name_for_port( TDSCONFIGINFO *config, TDSLOGIN *login );

extern int g_append_mode;

static char interf_file[MAXPATH];

/*
** tds_get_config() will fill the tds config structure based on configuration 
** information gathered in the following order:
** 1) Program specified in TDSLOGIN structure
** 2) The environment variables TDSVER, TDSDUMP, TDSPORT, TDSQUERY, TDSHOST
** 3) A config file with the following search order:
**    a) a readable file specified by environment variable FREETDSCONF
**    b) a readable file in ~/.freetds.conf
**    c) a readable file in $prefix/etc/freetds.conf
** 3) ~/.interfaces if exists
** 4) $SYBASE/interfaces if exists
** 5) TDS_DEF_* default values
**
** .tdsrc and freetds.conf have been added to make the package easier to 
** integration with various Linux and *BSD distributions.
*/ 
TDSCONFIGINFO *tds_get_config(TDSSOCKET *tds, TDSLOGIN *login, TDSLOCINFO *locale)
{
	TDSCONFIGINFO *config;
	char *s;
	char path[MAXPATH];
	pid_t pid;

	/* allocate a new structure with hard coded and build-time defaults */
locale->date_fmt = "%Y-%m-%d %H:%M:%S.%z";

	config = tds_alloc_config(locale);

	if (! tds_read_conf_file(login->server_name, config)) {
		/* fallback to interfaces file */
		tds_read_interfaces(login->server_name, config);
	}

	parse_server_name_for_port( config, login );

	tds_config_login(config, login);

	return config;
}        

static int tds_try_conf_file(char *path, char *how, char *server, TDSCONFIGINFO *config)
{
	int found = 0;
	FILE *in;

	config->major_version=7;
	config->minor_version=0;
	config->block_size = 512;
	return found;
}

static int tds_read_conf_file(char *server, TDSCONFIGINFO *config)
{
	int found = 0;

	tds_config_verstr("7.0", config);
	config->block_size = 512;
	return 0;
}
static int tds_config_boolean(char *value) 
{
	if (!strcmp(value, "yes") ||
		!strcmp(value, "on") ||
		!strcmp(value, "true") ||
		!strcmp(value, "1")) {
                return 1;
	} else {
                return 0;
	}
}

static void tds_read_interfaces(char *server, TDSCONFIGINFO *config)
{
	char ip_addr[255], ip_port[255], tds_ver[255];

	/* read $SYBASE/interfaces */
	/* This needs to be cleaned up */
	get_server_info(server, ip_addr, ip_port, tds_ver);
	if (strlen(ip_addr)) {
		if (config->ip_addr) free(config->ip_addr);
		/* FIXME check result, use strdup */
		config->ip_addr = (char *) malloc(strlen(ip_addr)+1);
		strcpy(config->ip_addr, ip_addr);
	}
	if (atoi(ip_port)) {
		config->port = atoi(ip_port);
	}
	if (strlen(tds_ver)) {
		tds_config_verstr(tds_ver, config);
		/* if it doesn't match a known version do nothing with it */
	}	
}

static void tds_config_login(TDSCONFIGINFO *config, TDSLOGIN *login)
{
	if (login->server_name && strlen(login->server_name)) {
		if (config->server_name) free(config->server_name);
		config->server_name = strdup(login->server_name);
	}	
	if (login->major_version || login->minor_version) {
		config->major_version = login->major_version;
		config->minor_version = login->minor_version;
	}
        if (login->language && strlen(login->language)) {
		if (config->language) free(config->language);
		config->language = strdup(login->language);
	}
        if (login->char_set && strlen(login->char_set)) {
		if (config->char_set) free(config->char_set);
		config->char_set = strdup(login->char_set);
	}
        if (login->host_name && strlen(login->host_name)) {
		if (config->host_name) free(config->host_name);
		config->host_name = strdup(login->host_name);
	}
        if (login->app_name && strlen(login->app_name)) {
		if (config->app_name) free(config->app_name);
		config->app_name = strdup(login->app_name);
	}
        if (login->user_name && strlen(login->user_name)) {
		if (config->user_name) free(config->user_name);
		config->user_name = strdup(login->user_name);
	}
        if (login->password && strlen(login->password)) {
		if (config->password) {
			/* for security reason clear memory */
			memset(config->password,0,strlen(config->password));
			free(config->password);
		}
		config->password = strdup(login->password);
	}
        if (login->library && strlen(login->library)) {
		if (config->library) free(config->library);
		config->library = strdup(login->library);
	}
        if (login->encrypted) {
		config->encrypted = 1;
	}
        if (login->suppress_language) {
		config->suppress_language = 1;
	}
        if (login->bulk_copy) {
		config->bulk_copy = 1;
	}
        if (login->block_size) {
		config->block_size = login->block_size;
	}
        if (login->port) {
		config->port = login->port;
	}

}

static void tds_config_verstr(char *tdsver, TDSCONFIGINFO *config)
{
	if (!strcmp(tdsver,"42") || !strcmp(tdsver,"4.2")) {
		config->major_version=4;
		config->minor_version=2;
		return;
	} else if (!strcmp(tdsver,"46") || !strcmp(tdsver,"4.6")) {
		config->major_version=4;
		config->minor_version=6;
		return;
	} else if (!strcmp(tdsver,"50") || !strcmp(tdsver,"5.0")) {
		config->major_version=5;
		config->minor_version=0;
		return;
	} else if (!strcmp(tdsver,"70") || !strcmp(tdsver,"7.0")) {
		config->major_version=7;
		config->minor_version=0;
		return;
	} else if (!strcmp(tdsver,"80") || !strcmp(tdsver,"8.0")) {
		config->major_version=8;
		config->minor_version=0;
		return;
	}
}


/* ============================== lookup_host() ==============================
 *
 * Def:  Given a servername and port name or number, lookup the
 *       hostname and service.  The server ip will be stored in the
 *       string 'servername' in dotted-decimal notation.  The service port
 *       number will be stored in string form in the 'port' parameter.
 *
 *       If we can't determine both the IP address and port number then
 *       'ip' and 'port' will be set to empty strings.
 *
 * Ret:  void
 *
 * ===========================================================================
 */
static void lookup_host(
   const char  *servername,  /* (I) name of the server                  */
   const char  *portname,    /* (I) name or number of the port          */
   char        *ip,          /* (O) dotted-decimal ip address of server */
   char        *port)        /* (O) port number of the service          */
{
   struct hostent   *host    = NULL;
   struct servent   *service = NULL;
   int               num     = 0;
   unsigned int		ip_addr=0;

   /* Storage for reentrant getaddrby* calls */
   struct hostent result;
   char buffer[4096];
   int h_errnop;

   /* Storage for reentrant getservbyname */
   struct servent serv_result;

   /* Only call gethostbyname if servername is not an ip address. 
      This call take a while and is useless for an ip address.
      mlilback 3/2/02 */
   ip_addr = inet_addr(servername);
   if (ip_addr == INADDR_NONE)
      host = tds_gethostbyname_r(servername, &result, buffer, sizeof(buffer), &h_errnop);

#ifndef NOREVERSELOOKUPS
/* froy@singleentry.com 12/21/2000 */
	if (host==NULL) {
		char addr [4];
		int a0, a1, a2, a3;
		sscanf (servername, "%d.%d.%d.%d", &a0, &a1, &a2, &a3);
		addr [0] = a0;
		addr [1] = a1;
		addr [2] = a2;
		addr [3] = a3;
		host    = tds_gethostbyaddr_r (addr, 4, AF_INET, &result, buffer, sizeof(buffer), &h_errnop);
	}
/* end froy */
#endif
   if (!host) {
      /* if servername is ip, copy to ip. */
      if (INADDR_NONE != ip_addr)
         strncpy(ip, servername, 17); 
      else
         ip[0]   = '\0';
   } else {
      struct in_addr *ptr = (struct in_addr *) host->h_addr;
      strncpy(ip, inet_ntoa(*ptr), 17);
   }
   if (portname) {
   	 service = tds_getservbyname_r(portname, "tcp", &serv_result, buffer, sizeof(buffer));
      if (service==NULL) {
         num = atoi(portname);
      } else {
         num = ntohs(service->s_port);
      }
   }

   if (num==0) {
	if (port) port[0] = '\0';
   } else {
	sprintf(port, "%d", num);
   }
} /* lookup_host()  */

static int hexdigit(char c)
{
	if (c>='a' && c<='f') {
		return c - 'a' + 10;
	} else if (c>='A' && c<='F') {
		return c - 'A' + 10;
	} else if (c>='0' && c<='9') {
		return c - '0';
	} else {
		return 0; /* bad hex digit */
	}
}
static int hex2num(char *hex)
{
	return hexdigit(hex[0])*16 + hexdigit(hex[1]);	
}


/* ============================ get_server_info() ============================
 *
 * Def:  Try to find the IP number and port for a (possibly) logical server
 *       name.
 *
 * Note: It is the callers responsibility to supply large enough buffers
 *       to hold the ip and port numbers.  ip_addr should be at least 17
 *       bytes long and ip_port should be at least 6 bytes long.
 *
 * Note: This function uses only the interfaces file and is deprecated.
 *
 * Ret:  True if it found the server, false otherwise.
 *
 * ===========================================================================
 */
int get_server_info(
   char *server,   /* (I) logical or physical server name      */
   char *ip_addr,  /* (O) string representation of IP address  */
   char *ip_port,  /* (O) string representation of port number */
   char *tds_ver)  /* (O) string value specifying which protocol version */
{
char  *sybase;
char  *home;
char  *tmp_port;
	ip_addr[0] = '\0';
	ip_port[0] = '\0';
	tds_ver[0] = '\0';

	if (!server || strlen(server) == 0) {
		server = getenv("TDSQUERY");
		if(!server || strlen(server) == 0) {
			server = "SYBASE";
		}
	}

	/*
	* if we haven't found the server yet then look for a $HOME/.interfaces file
	*/
	if (ip_addr[0]=='\0') {
		/* FIXME use getpwent, see above */
		home = getenv("HOME");
	}

	/*
	* if we haven't found the server yet then look in $SYBBASE/interfaces file
	*/
	if (ip_addr[0]=='\0') {
		sybase = getenv("SYBASE");
	}

 	/*
 	* If we still don't have the server and port then assume the user
 	* typed an actual server name.
 	*/
 	if (ip_addr[0]=='\0') {

		/*
		* Make a guess about the port number
		*/
#ifdef TDS50
		tmp_port = "4000";
#else
		tmp_port = "1433";
#endif

		/*
		* lookup the host and service
		*/
		lookup_host(server, tmp_port, ip_addr, ip_port);

	}

	return ip_addr[0]!='\0' && ip_port[0]!='\0';
} /* get_server_info()  */

/**
 * Check the server name to find port info first
 * return 1 when found, else 0
 * Warning: config-> & login-> are all modified when needed
 */
static int parse_server_name_for_port( TDSCONFIGINFO *config, TDSLOGIN *login )
{
    char *pSep, *pEnd;
        
    if( ! login->server_name )
        return 0;/* FALSE */
    
    /* seek the ':' in login server_name */
    pEnd = login->server_name + strlen( login->server_name );
    for( pSep = login->server_name; pSep < pEnd; pSep ++ )
        if( *pSep == ':' ) break;
    
    if(( pSep < pEnd )&&( pSep != login->server_name ))/* yes, i found it! */
    {
        if( config->server_name ) free( config->server_name );
        config->server_name = strdup( login->server_name );
        
        /* modify config-> && login->server_name & ->port */
        login->port = config->port = atoi( pSep + 1 );
        config->server_name[pSep - login->server_name] = 0;/* end the server_name before the ':' */
        *pSep = 0;

        /* config->ip_addr needed */
        {
            char tmp[256];

            lookup_host (config->server_name, NULL, tmp, NULL);
            if (config->ip_addr)
                    free (config->ip_addr);
            config->ip_addr = strdup (tmp);
        }

        return 1;/* TRUE */
    }
    else
        return 0;/* FALSE */
}

